# Common python helper functions used for Contrail charms.

import sys
import subprocess
import os, errno
import filecmp
import yaml
import re
import time
import tempfile
from datetime import datetime
from charmhelpers.core.hookenv import (
    log as juju_log,
    config,
    unit_private_ip,
    relation_ids,
    related_units,
    relation_get,
    relation_set,
    charm_dir,
)

dry_run = None

def fatal_cmd(cmd):
    subprocess.check_call(cmd)
#end fatal_cmd

def nonfatal_cmd(cmd):
    subprocess.call(cmd)
#end fatal_cmd

def mkdir(loc):
    try:
        os.makedirs(loc)
    except OSError as exc:
        if not (exc.errno == errno.EEXIST and os.path.isdir(loc)):
            raise BaseException('makedirs {} failed'.format(loc))
#end mkdir

def fatal_dircmd(loc, cmd):
    back = os.getcwd()
    try:
        os.chdir(loc)
        fatal_cmd(cmd)
    finally:
        os.chdir(back)
#end fatal_dircmd

def fatal_dir_nonfatal_cmd(loc, cmd):
    back = os.getcwd()
    try:
        os.chdir(loc)
        nonfatal_cmd(cmd)
    finally:
        os.chdir(back)
#end fatal_dircmd

def nonfatal_dircmd(loc, cmd):
    back = os.getcwd()
    try:
        os.chdir(loc)
        nonfatal_cmd(cmd)
    except:
        juju_log('chdir {} failure'.format(loc))
    finally:
        os.chdir(back)
#end fatal_dircmd

def relations(sections):
    rels = []
    with open(os.path.join(charm_dir(), 'metadata.yaml')) as mdf:
        md = yaml.safe_load(mdf)
        for s in sections:
            section = md.get(s)
            if section:
                rels.extend(section.keys())
    return rels
#end relations

def iface_inet(dev):
    import netifaces
    return netifaces.ifaddresses(dev)[netifaces.AF_INET][0]
#end iface_inet

def data_interface_details(dev):
    fatal_cmd(['ifup', dev])

    inet = iface_inet(dev)
    ip = inet['addr']
    gw = re.sub('\.255$', '.1', inet['broadcast'])

    return dev, ip, gw
#end data_interface_details

etc_dir = '/etc'
# TODO knob for data interface device, ip and gw
def data_interface():
    dev = 'eth1'
    if is_dry_run():
        juju_log("Dry run data_interface({})".format(dev))
        return dev, '2.2.2.2', '2.2.2.1'

    fname = os.path.join(etc_dir, 'network/interfaces')
    result = nonfatal_cmd(['grep', 'iface ' + dev, fname])
    if result == 0:
        juju_log("{} already present in {}".format(dev, fname))
        return data_interface_details(dev)

    tmpfname = tempfile.mktemp(dir='/tmp')
    with open(tmpfname, 'w') as tmpf:
        with open(fname, 'r') as inf:
            tmpf.write(inf.read())
        tmpf.write("auto %s\n" % dev)
        tmpf.write("iface %s inet dhcp\n" % dev)
    os.rename(tmpfname, fname)

    return data_interface_details(dev)
#end data_interface

def resolve(hn):
    #return socket.gethostbyname(unit_private_ip() ... localhost 127.
    cmd = ['dig', '+short', hn]
    out = subprocess.check_output(cmd)
    pat = re.compile("\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}")
    return pat.search(out).group()

def touch(fname):
    mkdir(os.path.dirname(fname))
    try:
        open(fname, 'w').close()
    except:
        juju_log("Could not touch {}".format(fname))
#end touch

def get_url(url, loc):
    # nc option ensures that file is not downloaded if it is already present
    cmd = ['wget', '-nc', '-nH', url]
    fatal_dircmd(loc, cmd)
#end get_url

def container_package_config():
    url = config('container-package')
    if not url: 
        raise BaseException('No URL specified')
    if url == 'dry-run':
        return None
    return url
#end container_package_config

def is_dry_run():
    global dry_run
    if dry_run == None:
        if container_package_config():
            dry_run = False
        else:
            dry_run = True
            juju_log("Dry run enabled")
    return dry_run
#end is_dry_run

container_dir = '/contrail-container-package'
def get_container_package():
    '''Get the container package from the URL specified in config'''

    url = container_package_config()
    if not url:
        return
    juju_log('container-package {}'.format(url))
    mkdir(container_dir)

    # Avoid multiple copies with LXCs
    if container_dir == '/contrail-container-package':
        container = os.path.basename(config('container-package'))
        if os.path.exists('/var/log/juju/' + container):
            fatal_dir_nonfatal_cmd(container_dir, 
                         ['ln', '-s', '/var/log/juju/' + container])

    get_url(url, container_dir)
#end get_container_package

already_dir = '/tmp/already/'
def unpack_container_package():
    '''Unpack the container package specified in config'''

    unpacked = already_dir + 'unpacked'
    if os.path.exists(unpacked):
        juju_log("{} exists - skip unpacking".format(unpacked))
        return

    container = os.path.basename(config('container-package'))
    cmd = ['dpkg', '-i', container]
    fatal_dircmd(container_dir, cmd)
    touch(unpacked)
#end unpack_container_package

# Temporary workaround for "12.04.4 LTS, Precise Pangolin" and/or
# container package not having correct package versions
def fix_it_up(cmd, loc='.'):
    fatal_dir_nonfatal_cmd(loc, cmd)
    #nonfatal_dircmd(loc, cmd)

def patch_loc(bname):
    return os.path.join(
        charm_dir(), 'hooks/charmhelpers/contrib/contrail/patches/' + bname)
#end patch_loc

def patch_it_up():
    if is_dry_run():
        juju_log("Dry run patch_it_up")
        return
    prefix = '/opt/contrail/'
    
    patched = already_dir + 'patched'
    if os.path.exists(patched):
        juju_log("{} exists - skip patching".format(patched))
        return

    for fname in [
        #'contrail_installer/contrail_setup_utils/setup-quantum-in-keystone.py',
        'contrail_installer/contrail_setup_utils/setup.py', # setup_vrouter
        'utils/fabfile/tasks/provision.py']:
        bname = os.path.basename(fname)
        cmd = ['patch', '-i', patch_loc(bname), bname]
        fatal_dircmd(prefix + os.path.dirname(fname), cmd)

    touch(patched)
#end patch_it_up
        
def create_install_repo():
    created = '/opt/contrail/utils'
    if os.path.exists(created):
        juju_log("{} exists - skip create install repo".format(created))
        return

    cmd = ['./setup.sh']
    fatal_dircmd('/opt/contrail/contrail_packages', cmd)
#end create_install_repo

def fabloc(loc, cmd, TESTBED=None):
    cmd.insert(0, 'fab')
    if TESTBED:
        cmd.insert(0, 'TESTBED=' + TESTBED)
        cmd.insert(0, 'sudo')
    if is_dry_run():
        juju_log("Dry run {}".format(cmd))
        return
    fatal_dircmd(loc, cmd)
#end fab

def fab(cmd, TESTBED):
    fabloc('/opt/contrail/utils', cmd, TESTBED)
#end fab

# Not called. Chicken-and-egg situation
#    'fab' needs 'testbed.py' which needs repo to be installed
def create_install_repo_fab(TESTBED):
    fab(['create_install_repo'], TESTBED)
#end create_install_repo_fab

def testbed_for_TESTBED():
    fname = testbed_dir + 'testbed.py'
    with open(fname, 'w') as f:
        f.write("import os\n")
        f.write("if os.environ.has_key('TESTBED'):\n")
        f.write("  execfile(os.environ['TESTBED'])\n")
        f.write("else:\n")
        f.write("  print 'TESTBED not set in environment'\n")
        f.write("  exit(1)\n")
    return fname

testbed_dir = '/opt/contrail/utils/fabfile/testbeds/'
def TESTBED_file(role, action):
    return testbed_dir + 'testbed_for_' + action + '_' + role + '.py'

# The charm will run as 'root', but since there is no password for 'root',
# we need SSH to be password-less
def passwordless_ssh():
    keydir = '/root/.ssh/'
    keyfile = keydir + 'id_rsa'
    if not os.path.exists(keyfile):
        cmd = ['sudo', 'ssh-keygen', '-P', '', '-f', keyfile]
        fatal_cmd(cmd)
    authfile = keydir + 'authorized_keys2'
    cmd = ['sudo', 'cp', keyfile + '.pub', authfile]
    fatal_cmd(cmd)

def install_packages_from_container():
    if is_dry_run():
        global testbed_dir
        mkdir(testbed_dir)
        juju_log("Dry run install_packages_from_container")
        return
    unpack_container_package()
    create_install_repo()
#end install_packages_from_container

def unit_ip(rel, r, u):
    hn = relation_get('private-address', rid=r, unit=u)
    ip = resolve(hn)
    juju_log('{} u: {} host: {} ip: {}'.format(rel, u, hn, ip))
    return ip
#end unit_ip

def relation_roleips(roles):
    #d = dict.fromkeys(roles, []) ... yielded unexpected results
    d = {role:[] for role in roles}
    for role in d:
        rel = role
        for r in relation_ids(rel):
            for u in related_units(r):
                ip = unit_ip(rel, r, u)
                d[role].append(ip)
        #d[role].sort() # MGTODO revisit?
    return d
#end relation_roleips

def remote_roleips(roles):
    d = {role:[] for role in roles}
    rels = relations(['provides'])
    for rel in rels:
        for r in relation_ids(rel):
            for u in related_units(r):
                role = relation_get('role', rid=r, unit=u)
                if not role in roles:
                    continue
                ip = unit_ip(rel, r, u)
                d[role].append(ip)
    #for role in d:
        #d[role].sort() # MGTODO revisit?
    return d
#end remote_roleips

def relation_jamming(role):
    rel = role
    for r in relation_ids(rel):
        for u in related_units(r):
            jamming = relation_get('jamming', rid=r, unit=u)
            juju_log('{} u: {} jamming: {}'.format(rel, u, jamming))
            if jamming == 'no' or not jamming:
                return False
    return True
#end relation_jamming

def relation_roledefs(roles, remote=False):
    if remote:
        roleips = remote_roleips(roles.keys())
    else:
        roleips = relation_roleips(roles.keys())
    for role in roleips:
        roledef = roles[role].get('roledef')
        if len(roleips[role]) == 0:
            if roledef != 'dummy':
                #TODO figure out cluster size
                juju_log('Awaiting IP address {}'.format(role))
                return {}
            else:
                juju_log('Dummy IP address {}'.format(role))
                #roleips[role].append('1.1.1.1')
                roleips[role].append('127.0.0.1')
        elif roledef == 'jamming':
            if not relation_jamming(role):
                juju_log('Awaiting jamming {}'.format(role))
                return {}
    juju_log('Quorum for env.roledefs {}'.format(roleips))
    return roleips
#end relation_roledefs

def roledefs(roleips, fname):
    with open(fname, 'a') as f:
        for role in roleips:
            f.write("env.roledefs['%s'] = [%s]\n" % 
                    (role, ', '.join(['\'root@' + ip + '\'' 
                                      for ip in roleips[role]])))
#end roledefs

def knob_attrib(knobs, fname):
    with open(fname, 'a') as f:
        for k in knobs:
            f.write("%s = %s\n" % (k, knobs[k]))
#end knob_attrib

def data_attrib(fname):
    with open(fname, 'a') as f:
        f.write("data = {host1: {'device': '%s', 'ip': '%s', 'gw': '%s'}}\n" %
                data_interface())
#end data_attrib

# Not called
def verify_jamming(role):
    TESTBED = TESTBED_file(role, 'start')
    try:
        jamming = 'no'
        task = 'verify_' + role
        fab([task], TESTBED)
    except:
        juju_log('{} failure - we not jamming'.format(task))
    else:
        juju_log('{} success - we be jamming!'.format(task))
        jamming = 'yes'
    return jamming
#end verify_jamming

def role_state_loc(fname=''):
    return os.path.join(charm_dir(), 'role/' + fname)
#end role_state_loc

def clean_start():
    cmd = ['rm', '-rf', role_state_loc()]
    nonfatal_cmd(cmd)
    mkdir(role_state_loc())
    juju_log("role_state now clean {}".format(role_state_loc()))
#end clean_start

def pre_start(TESTBED):
    ''' clean -> starting or started -> starting or jamming -> started '''

    new = 'starting'
    new = role_state_loc(new)

    old = 'jamming'
    old = role_state_loc(old)
    try:
        os.rename(old, new)
    except:
        try:
            old = 'started'
            old = role_state_loc(old)
        except:
            open(new, 'a').close()

    with open(new, 'a') as outf:
        outf.write("\nvvv %s %s\n" % (str(datetime.now()), TESTBED))
        if TESTBED:
            with open(TESTBED, 'r') as inf:
                outf.write(inf.read()) # TESTBED >> starting
    juju_log("role_state {} -> {}".format(os.path.basename(old), 
                                          os.path.basename(new)))
#end pre_start

def post_start(jam='later'):
    ''' starting -> started or starting -> jamming or started -> jamming '''

    if jam == 'later':
        old = 'starting'
        new = 'started'
    elif jam == 'immediate':
        old = 'starting'
        new = 'jamming'
    elif jam == 'finally':
        old = 'started'
        new = 'jamming'
    else:
        raise BaseException('unknown jam option')
    old = role_state_loc(old)
    new = role_state_loc(new)
    os.rename(old, new)
    if jam != 'later':
        with open(new, 'a') as outf:
            outf.write("\n^^^ %s\n" % str(datetime.now()))
    juju_log("role_state {} -> {}".format(os.path.basename(old), 
                                          os.path.basename(new)))
#end post_start

def be_jamming(role):
    fname = role_state_loc('jamming')
    if not os.path.exists(fname):
        juju_log('{} absent - we not jamming'.format(fname))
        jamming = 'no'
    else:
        juju_log('{} present - we be jamming'.format(fname))
        jamming = 'yes'
    return jamming
#end be_jamming

def role_state():
    states = ['jamming', 'started', 'starting']
    for s in states:
        fname = role_state_loc(s)
        if os.path.exists(fname): 
            return s
    return 'clean'
#end role_state

# See https://bugs.launchpad.net/juju-core/+bug/1244761
# Workaround: Do our very best not to cause reboot mid-hook
def reboot(skip=True):
    if is_dry_run():
        skip = True

    child_pid = os.fork()
    if child_pid == 0:
        # child process
        if skip:
            sys.exit(os.execl('/bin/echo', 'echo', 'Child skip reboot'))

        time.sleep(10)
        # TODO fab compute_reboot
        sys.exit(os.execl('/sbin/reboot', 'reboot'))
    else:
        # parent process
        # pid, status = os.waitpid(child_pid, 0)
        juju_log("Reboot scheduled (skip={}) via child pid {}".format(
                skip, child_pid))
        return skip
#end reboot

class Role(object):
    def __init__(self, role, also=[], ifabs=[],
                 reqd={}, knobs={}, sfabs=[], pfabs=[],
                 adj={}, afabs=[], jam=False):
        self.role = role
        self.also = also
        self.ifabs = ifabs
        self.reqd = reqd
        self.knobs = knobs
        self.sfabs = sfabs
        self.pfabs = pfabs
        self.adj = adj
        self.afabs = afabs
        self.jam = jam
    #end __init__

    def jamming(self, jamming='determine'):
        '''Pertains to all services provided by this unit'''

        if not self.jam:
            return

        if jamming == 'determine':
            jamming = be_jamming(self.role)

        rels = relations(['provides'])
        for rel in rels:
            for r in relation_ids(rel):
                # https://bugs.launchpad.net/juju-core/+bug/1239681
                # current = relation_get('jamming', rid=r)
                # if current == jamming:
                #    continue
                relation_set(r, {'jamming': jamming})
        return rels
    #end jamming

    def _testbed_for_install(self):
        fname = TESTBED_file(self.role, 'install')
        with open(fname, 'w') as f:
            f.write("from fabric.api import env\n")

            # Need this because...
            #   host_string.split('@')[1]
            #   cassandra seeds: comma-delimited list of addresses 
            #   cfgm: IP address (not hostname)
            # f.write("host1 = 'localhost'\n") is sufficient otherwise
            f.write("host1 = 'root@%s'\n" % resolve(unit_private_ip()))

            # Need this because...
            #   database_host_password=env.passwords[host_string]
            f.write("env.passwords = {host1: ''}\n")

            f.write("env.roledefs = {'all': [host1],\n")
            f.write("                'build':  [host1],\n")
            f.write("                '%s': [host1]}\n" % self.role)

            for r in self.also:
                f.write("env.roledefs['%s'] = [host1] # overridden in testbed_for_start\n" % r)

            return fname
    #end _testbed_for_install

    def _testbed_for_start(self, compare):
        roleips = {}
        if self.reqd:
            roleips = relation_roledefs(self.reqd)
            if not roleips:
                return None # Not ready

        fname = TESTBED_file(self.role, 'start')

        # Save the previous testbed for comparison with new
        if compare:
            prev = fname + '.prev'
            try:
                os.rename(fname, prev)
            except:
                prev = None

        # Construct the new file
        head = TESTBED_file(self.role, 'install')
        if not os.path.exists(head):
            head = self._testbed_for_install()
        with open(fname, 'w') as outf:
            with open(head, 'r') as inf:
                outf.write(inf.read()) # install > start
        roledefs(roleips, fname)
        if self.knobs:
            knob_attrib(self.knobs, fname)
        if self.role == 'compute':
            # Typically, MAAS+Juju do not provide login access via console. 
            # If the management interface is hosed, it get difficult to debug 
            # Data interface allows the management interface to be left alone.
            data_attrib(fname)

        # Compare
        if compare and prev and filecmp.cmp(prev, fname):
            juju_log("No change in {}".format(fname))
            return '' # skip

        return fname # ready
    #end _testbed_for_start

    def _testbed_for_adjust(self):
        roleips = {}
        if self.adj:
            roleips = relation_roledefs(self.adj, remote=True)
            if not roleips:
                return None # not ready

        fname = TESTBED_file(self.role, 'adjust')

        # Construct 
        head = TESTBED_file(self.role, 'start')
        if not os.path.exists(head):
            head = self._testbed_for_start(compare=False)
        with open(fname, 'w') as outf:
            with open(head, 'r') as inf:
                outf.write(inf.read()) # start > adjust

        roledefs(roleips, fname)

        return fname # ready (may be a repeat)
    #end _testbed_for_start

    # Only one role per charm...
    # See fabfile/tasks/install.py:install_contrail()
    # install_webui and install_collector needs install_cfgm
    # install_cfgm needs install_collector (pycassa)
    def _install(self):
        testbed_for_TESTBED()
        TESTBED = self._testbed_for_install()

        # Also install the roles as a workaround for missing packages
        for r in self.also:
            installed = already_dir + 'installed_' + r
            if os.path.exists(installed):
                continue
            fab(['install_' + r], TESTBED)
            touch(installed)

        if self.ifabs:
            for f in self.ifabs:
                fab([f], TESTBED)
        else:
            installed = already_dir + 'installed_' + self.role
            if not os.path.exists(installed):
                fab(['install_' + self.role], TESTBED)
                fab(['upgrade_pkgs'], TESTBED)
            touch(installed)
        
    #end _install

    def _start(self):
        rs = role_state()
        juju_log("_start: role_state is {}".format(rs))

        if rs == 'started':
            # get out of here ASAP, reboot a-coming!
            # if reboot catches us mid-hook, charm will end up in error state
            return 'started'

        compare = (rs == 'jamming')

        TESTBED = self._testbed_for_start(compare)
        if not TESTBED:
            self.jamming('yes')
            return True # already

        self.jamming('no')

        if not TESTBED:
            return False # awaiting quorum

        pre_start(TESTBED)

        if self.sfabs:
            for f in self.sfabs:
                fab([f], TESTBED)
        else:
            fab(['setup_' + self.role], TESTBED)
            fab(['verify_' + self.role], TESTBED)
        # fabric.network.connections.clear() - MGTODO understand

        if self.pfabs:
            post_start('later')
            reboot()
            return False # reboot a-coming
        else:
            post_start('immediate')
            self.jamming('yes')
            return True # done
    #end _start

    def _proceed(self):
        rs = role_state()
        juju_log("_proceed: role_state is {}".format(rs))

        if rs == 'jamming':
            return True

        if rs != 'started':
            return False

        for f in self.pfabs:
            fab([f], TESTBED_file(self.role, 'start'))

        post_start('finally')
        self.jamming('yes')
        return True
    #end _proceed

    def _adjust(self):
        rs = role_state()
        juju_log("_adjust: role_state is {}".format(rs))

        if rs != 'jamming':
            return False # TODO revisit this

        TESTBED = self._testbed_for_adjust()
        if not TESTBED:
            return False # awaiting quorum

        for f in self.afabs:
            fab([f], TESTBED)

        return True # done
    #end _adjust

    def install(self):
        if self.role == 'compute' and os.uname()[2] != '3.8.0-29-generic':
            fix_it_up(['sudo', 'DEBIAN_FRONTEND=noninteractive', 
                       'apt-get', '-y', 'install', 
                       'linux-image-3.8.0-29-generic'])
        get_container_package()
        install_packages_from_container()
        passwordless_ssh()
        self._install()
        patch_it_up()
    #end install

    def install_all(self):
        if os.uname()[2] != '3.8.0-29-generic':
            fix_it_up(['sudo', 'DEBIAN_FRONTEND=noninteractive', 
                       'apt-get', '-y', 'install', 
                       'linux-image-3.8.0-29-generic'])
        get_container_package()
        install_packages_from_container()
        passwordless_ssh()
        fname = testbed_dir + 'testbed.py'
        with open(fname, 'w') as f:
            f.write("from fabric.api import env\n")
            f.write("host1 = 'root@%s'\n" % resolve(unit_private_ip()))
            f.write("env.passwords = {host1: ''}\n")
            f.write("env.roledefs = {'all': [host1],\n")
            for r in ['cfgm', 'openstack', 'control', 'compute', 'collector', 
                      'webui', 'database']:
                f.write("                '%s': [host1],\n" % r)
            f.write("                'build':  [host1]}\n")
        knob_attrib({'ext_routers': [],
                     'router_asn': 64512,
                     'env.interface_rename': False,
                     'service_token': "'contrail123'"}, fname)
        fab(['install_contrail'], fname)
    #end install_all

    def start(self, clean=False):
        if clean:
            clean_start()

        return self._start()
    #end start

    def proceed(self):
        if not self.pfabs:
            return True

        return self._proceed()
    #end proceed

    def adjust(self):
        if not self.afabs or not self.adj:
            return True

        return self._adjust()
    #end adjust

    def join(self):
        rels = relations(['provides', 'requires'])
        for rel in rels:
            for r in relation_ids(rel):
                relation_set(r, {'role': self.role})
        return rels
    #end join

#end Role
